//////////////////////////////////////////////////////////////////////////////
//
// File:	neural.h
// Directory:	/users/hartmann/rapid/src
//
//////////////////////////////////////////////////////////////////////////////

#ifndef NEURAL_H
#define NEURAL_H

#include "black.h"
#include <math.h>


// boxes--type identifiers in neural.C
extern const char *LogisticBP0;
extern const char *Logistic;
extern const char *LogisticBP;
extern const char *BipSigmoid;
extern const char *BipSigmoidBP;
extern const char *BipSigmoidRBP;
extern const char *BipSigmoidBBP;
extern const char *Synapses;
extern const char *SynapsesBP;
extern const char *SynapsesMBP;
extern const char *SynapsesBBP;
extern const char *TeacherE;
extern const char *TeacherWE;
extern const char *TeacherBE;
extern const char *Confusion;
extern const char *Energizer;
extern const char *EnergyE;

//key words
extern const char *InFNameKey;
extern const char *OutFNameKey;
extern const char *FormatKey;
extern const char *AsciiKey;
extern const char *BinKey;
extern const char *FillKey;
extern const char *TeeFNameKey;
extern const char *MaximKey;
extern const char *SlopeKey;
extern const char *LRateKey;
extern const char *LDecayKey;
extern const char *MaxRndKey;
extern const char *WDecayKey;
extern const char *WeightKey;
extern const char *ThldKey;



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: logisticBP0
//
//////////////////////////////////////////////////////////////////////////////

// logistic function convertor - the basic version
//   -in backward track the error is propagated backwards
//     Yb signals are multiplied by the derivative
//     of log.function at points given by X (i.e. Youts) signals
// !!! important !!!: Maximum and Slope are supposed to be 1 and not adapted 

class logisticBP0 : public blackBox
{
   
   protected:
     signal 	*X, *Y;
     signal	*Xb, *Yb;
     int	N;
            
   public:
     logisticBP0
     (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual sysig forward();
      virtual sysig backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: logistic
//
//////////////////////////////////////////////////////////////////////////////

// logistic function convertor
//   -non-linear limitation using sigmoidal logistic function in forward track
//   -no connections in backward track
//
// if Maximum or Slope not specified, the values 1.0 or 1.0 are assumed

class logistic : public blackBox
{
   
   protected:
      double	MaximumVal, SlopeVal;
      double	*Maximum, *Slope;
      double 	*X, *Y;
      int	N;
            
   public:
     logistic
     (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      
      virtual sysig forward();
      virtual sysig backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: logisticBP
//
//////////////////////////////////////////////////////////////////////////////

// logistic function convertor
//   -forward track inherited from logistic
//   -in backward track the error is propagated backwards
//     Yb signals are multiplied by the derivative
//     of log.function at points given by X (i.e. Youts) signals
//   -if LearnRate != 0.0 then slope adaptation alowed

class logisticBP : public logistic
{
   
   protected:
     double	*Xb, *Yb;
     double	LearnRate;
            
   public:
     logisticBP
     (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: logistic(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig  backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: bipsigmoid
//
//////////////////////////////////////////////////////////////////////////////

// bipolar sigmoidal non-linearity
// Maximum, Slope are common to all processing units

class bipsigmoid : public blackBox
{
   
   protected:
      double	MaximumVal, SlopeVal;
      double	*Maximum, *Slope;
      double 	*X, *Y;
      int	N;
            
   public:
      bipsigmoid
     (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig forward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: bipsigmoidBP
//
//////////////////////////////////////////////////////////////////////////////

// bipolar sigmoidal non-linearity, adaptiv and backpropagating error

class bipsigmoidBP : public bipsigmoid
{
   
   protected:
     double	*Xb, *Yb;
     double	LearnRate;
            
   public:
      bipsigmoidBP
     (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: bipsigmoid(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig  backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: bipsigmoidRBP
//
//////////////////////////////////////////////////////////////////////////////

// bipolar sigmoidal non-linearity, adaptiv and backpropagating error "roughly"
//  "roughly" means that y(1-y) factor is ignored

class bipsigmoidRBP : public bipsigmoidBP
{

   public:
      bipsigmoidRBP
        (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: bipsigmoidBP(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual sysig  backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: bipsigmoidBBP
//
//////////////////////////////////////////////////////////////////////////////

// bipolar sigmoidal non-linearity, backpropagating error and adaptiv
//  if Ctrl signals are non-zero (-->batch mode)

class bipsigmoidBBP : public bipsigmoidBP
{

   protected:      
      signal	*CtrlIn, *CtrlOut;

   public:
      bipsigmoidBBP
        (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: bipsigmoidBP(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:
      virtual result init();
      virtual sysig  backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: synapses
//
//////////////////////////////////////////////////////////////////////////////

// 'synapses' has 1 input and 1 output connection (vector) in forward track
//   -it computes the synaptic potencial based on weights and thresholds
//   -no backward algorithm (adaptation) is defined
//

class synapses : public blackBox
{
   
   protected:
      int	Nx, Nv;
      signal	*X, *V;
      double	**W;	// indexing weights: "1 output neuron per line"
      			// W<output (row) index k,input (column) index j>
      			//   = W[k][j]
      			//  i.e. #rows:	Nv	#columns: Nx
      			// matrix form: V = W * X
      
      double	*Thld;	// field of thresholds
      
   public:
     synapses(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      
      virtual sysig forward();
      virtual sysig backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: synapsesBP
//
//////////////////////////////////////////////////////////////////////////////

// 'synapsesBP' has 1 connection on all the tracks | inputs | outputs
//   -it computes the synaptic potencial based on weights and thresholds
//     as inherited from synapses
//   -backward propagation of the gradient as well as adaptation is possible
//     in the basic form of backpropagation algorithm
//

class synapsesBP : public synapses
{
   
   protected:
      double	RandMax,LearnRate;
      signal	*Xb, *Vb;

      
   public:
     synapsesBP(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: synapses(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: synapsesMBP
//
//////////////////////////////////////////////////////////////////////////////

// back-propagation with "momentum"

class synapsesMBP : public synapsesBP
{
   
   protected:
      double 	**ddW, *ddThld;
      double 	Decay;
      
   public:
     synapsesMBP(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: synapsesBP(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: synapsesBBP
//
//////////////////////////////////////////////////////////////////////////////

// synaptic layer, backpropagating error and adaptiv
//  if Ctrl signals are non-zero (-->batch mode)

class synapsesBBP : public synapsesBP
{

   protected:      
      signal	*CtrlIn, *CtrlOut;

   public:
      synapsesBBP
        (char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: synapsesBP(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:
      virtual result init();
      virtual sysig  backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: teacherE
//
//////////////////////////////////////////////////////////////////////////////

// teacherE assumes desired and actual inputs (0. and 1. port) at the X side,
//   error vector = desired - actual at the X output
//   and poosibly euclidian distance at the Y output
// forward method computes error vector and returns Bsig
//   or error vector and distance and returns Fsig
// backward method computes error vector and returns Bsig

class teacherE : public blackBox
{
   
   protected:
      signal	*X, *d, *e, *dist;
      int 	N;

      
   public:
      teacherE(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual sysig forward();
      virtual sysig backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: teacherWE
//
//////////////////////////////////////////////////////////////////////////////

// teacherWE is like teacherE 
//  but each element of error vector e[i] is premultiplied by some coeffitient

class teacherWE : public teacherE
{
   
   protected:
       signal	*k;
      
   public:
      teacherWE(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: teacherE(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig forward();
      virtual sysig backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: teacherBE
//
//////////////////////////////////////////////////////////////////////////////

// teacherBE is like teacherE but accumulates the error vector as well as
// the distance 
//   if number of vectors summed up achieves the BlockSize parameter
//   the CtrlDist signal is filled by distance (otherwise zero)
//   thus the neural network can be trained in the batch-like mode
//   if all adaptive layers check the CtrlDist signal.

class teacherBE : public teacherE
{
      
   protected:      
      int	C, BlockSize;
      signal	*CtrlDist;
      signal	EDist;
      
   public:
      teacherBE(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: teacherE(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      
      virtual sysig forward();
      virtual sysig backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: confusion
//
//////////////////////////////////////////////////////////////////////////////

class confusion : public blackBox
{
   
   protected:
      int	X, D;
      int 	N;
      int	**M;
      
      FILE	*ConfTabFile, *ConfMatFile;
      char	*ConfTabName, *ConfMatName;
      
   public:
     confusion(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig forward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: energizer
//
//////////////////////////////////////////////////////////////////////////////

class energizer : public blackBox
{
   
   protected:
      long int 	n,N;
      signal	S;
      
   public:
     energizer(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig forward();
      virtual sysig backward();

};



//////////////////////////////////////////////////////////////////////////////
//
// CLASS: energyE
//
//////////////////////////////////////////////////////////////////////////////

// sums up the squares of differences of the incomming signal
//  at the end writes the total on stdout

class energyE : public blackBox
{
   
   protected:
      signal	*X, *Y, E;
      int 	N;
      
   public:
      energyE(char *TypeStr,char *SignatStr,int Xins,int Youts,int Yins,int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	 
   protected:      
      virtual result init();
      virtual sysig forward();
      virtual result shutdown();

};



#endif
