///////////////////////////////////////////////////////////////////////////
//
// File:	neural2.C
//
// Directory:	/users/pbienst/mlp
//
///////////////////////////////////////////////////////////////////////////


#include "neural.h"
#include "neural2.h"

// boxes--type identifiers

const char *SynapsesTD     = "synapsesTD";
const char *SynapsesTDBP   = "synapsesTDBP";
const char *SynapsesTD2    = "synapsesTD2";
const char *SynapsesTDBP2  = "synapsesTDBP2";
const char *SoftMax        = "softmax";
const char *SoftMaxBP      = "softmaxBP";
const char *Teacher        = "teacher";
const char *Tester         = "tester";
const char *ConfusionDummy = "confusiondummy";

// key words

const char *DelaysKey = "Delays";

//// synapsesTD::init /////////////////////////////////
//

result synapsesTD::init()
{
   
   // reading number of delays
   
   if(ReadVal((char*)DelaysKey,Delays) != OKAY)
   {
      ErrOutput << Sign << DelaysKey << SepEQU << " not specified.\n";
      return FAILED;
   }
  
   // input/output locations and sizes
   
   int i,t;
   
   Nx = Xin->Size[0];
   AllocArray(X, Delays+1, Nx);
   for (t=0; t<=Delays; t++)
      for (i=0; i<Nx; i++)
	 X[t][i]=0;
 
 
   V  = Yout->Data[0];
   Nv = Yout->Size[0];

   
   // create weights and thresholds
   
   W = NULL;   Thld = NULL;
   AllocArray(W, Delays+1, Nv, Nx);
   AllocArray(Thld, Delays+1, Nv);
   if(!W || !Thld)
      return FAILED;
      
   // load weights and thresholds

   if(ReadArray((char*)WeightKey, W, Delays+1, Nv, Nx) != OKAY)
   {
      ErrOutput << Sign << "No 3-d weight matrix specified.\n";
      return FAILED;
   }

   if(ReadArray((char*)ThldKey, Thld, Delays+1, Nv) != OKAY)
   {
      ErrOutput << Sign << "No 2-d matrix of thresholds specified.\n";
      return FAILED;
   }

 
   return OKAY;
   
};



//// synapsesTD::save /////////////////////////////////
//

result synapsesTD::save()
{
   
   WriteVal((char*)DelaysKey,Delays);WriteLn();
   WriteArray((char*)WeightKey, W, Delays+1, Nv, Nx);
   WriteArray((char*)ThldKey, Thld, Delays+1, Nv);

   return OKAY;

};   

//// synapsesTD::forward /////////////////////////////////
//

sysig synapsesTD::forward()
{

   int t,k,j;
   
   // reading new input
   
   for (j=0; j<Nx; j++)
      X[0][j] = Xin->Data[0][j];

   
   // calculating output
   
   for(k=0;k<Nv;k++)
   {
      V[k]=0.0;
      for (t=0; t <= Delays; t++)
      {
         for(j=0;j<Nx;j++)
            V[k] += W[t][k][j] * X[t][j];
         V[k] -= Thld[t][k];
      }
   }

   // shifting input buffer
   
   for (t=1; t<=Delays; t++)
      for (j=0; j<Nx; j++)
	 X[t][j] = X[t-1][j];
 
    
   return Fsig;

};

//// synapsesTD::backward /////////////////////////////////
//

sysig synapsesTD::backward()
{

   return NOsig;

};

//// synapsesTDBP::init /////////////////////////////////
//

result synapsesTDBP::init()
{
   
   int t,k,j;
   
   // reading number of delays
   if(ReadVal((char*)DelaysKey,Delays) != OKAY)
   {
      ErrOutput << Sign << DelaysKey << SepEQU << " not specified.\n";
      return FAILED;
   }

   // input/output locations and sizes
   
   Nx = Xin->Size[0];
   AllocArray(X, Delays+1, Nx);
   for (t=0; t<=Delays; t++)
      for (j=0; j<Nx; j++)
	 X[t][j]=0;
   Xb = Xout->Data[0];
  
   if(Xout->Size[0])
      if(Nx != Xout->Size[0])
      {
         ErrOutput << Sign << "Vectors X and Xb have diferent sizes.\n";
         return FAILED;
      }
 
   V = Yout->Data[0];
   Vb = Yin->Data[0];
   if((Nv = Yout->Size[0]) != Yin->Size[0])
   {
      ErrOutput << Sign << "Vectors V and Vb have diferent sizes.\n";
      return FAILED;
   }

   
   // reading learning rate
   if(ReadVal((char*)LRateKey,LearnRate) != OKAY)
   {
      ErrOutput << Sign << LRateKey << SepEQU << " not specified.\n";
      return FAILED;
   }
      
   // create weights and thresholds
   W = NULL;   Thld = NULL;
   AllocArray(W, Delays+1, Nv, Nx);
   AllocArray(Thld, Delays+1, Nv);
   if(!W || !Thld)
      return FAILED;

   // set weights, if not given then zeros
   boolean ZeroWeights = FALSE;
   if(ReadArray((char*)WeightKey, W, Delays+1, Nv, Nx) != OKAY)
   {
      for (t=0; t<=Delays; t++) 
         for(k=0;k<Nv;k++)
            for(j=0;j<Nx;j++)
	       W[t][k][j] = 0.0;
      ZeroWeights = TRUE;
   }
   
   // set thresholds, if not given then zeros
   if(ReadArray((char*)ThldKey, Thld, Delays+1, Nv) != OKAY)
      for (t=0; t<=Delays; t++)
         for(k=0;k<Nv;k++)
     	    Thld[t][k] = 0.0;
   	
   // if some random bound is given
   //   add noise to the weights&thresholds
   // in this way the weights&thresholds can be preset in random manner
   if(ReadVal((char*)MaxRndKey, RandMax) == OKAY)
   { 
      // adding noise, DC = 0
      for (t=0; t<=Delays; t++)
         for(k=0;k<Nv;k++)
         {
	    for(j=0;j<Nx;j++)
	       W[t][k][j] = 2.0*RandMax * randdouble - RandMax;
     	    Thld[t][k] = 2.0*RandMax * randdouble - RandMax;
         }
      }

   else
      if(ZeroWeights)
         WarnOutput << Sign << "All weights are set to zero.\n";
      

   return OKAY;
      
};




//// synapsesTDBP::save /////////////////////////////////
//

result synapsesTDBP::save()
{
   
   WriteVal((char*)DelaysKey,Delays);WriteLn();
   WriteVal((char*)LRateKey,LearnRate);WriteLn();
   WriteArray((char*)WeightKey, W, Delays+1, Nv, Nx);
   WriteArray((char*)ThldKey, Thld, Delays+1, Nv);

   return OKAY;

}; 

//// synapsesTDBP::forward /////////////////////////////////
//

sysig synapsesTDBP::forward()
{

   int t,k,j;
   
   // reading new input

   for (j=0; j<Nx; j++)
      X[0][j] = Xin->Data[0][j];

   
   // calculating output
   
   for(k=0;k<Nv;k++)
   {
      V[k]=0.0;
      for (t=0; t <= Delays; t++)
      {
	 for(j=0;j<Nx;j++)
            V[k] += W[t][k][j] * X[t][j];
         V[k] -= Thld[t][k];
      }
   }
   
   // shifting input buffer should be delayed until after
   // backward function
   
      
   return Fsig;

};

//// synapsesTDBP::backward /////////////////////////////////
//

sysig synapsesTDBP::backward()
{

   int t,k,j;
   
   
   // backpropagation first, only for zero delay layer

   if(Xout->Size[0])
     for(j=0;j<Nx;j++)
     {
	Xb[j] = 0.0;
	for(k=0;k<Nv;k++)
	   Xb[j] += W[0][k][j] * Vb[k];
     }
   
     
   // adaptation
   
   if(LearnRate != 0.0)
      for (t=0; t<=Delays; t++)
         for(k=0;k<Nv;k++)
         {
	    for(j=0;j<Nx;j++)
	       W[t][k][j] += LearnRate * Vb[k] * X[t][j];
            Thld[t][k] -= LearnRate * Vb[k];
         }
   
   // now it's safe to shift the buffer
   
   for (t=1; t<=Delays; t++)
      for (j=0; j<Nx; j++)
	 X[t][j] = X[t-1][j];
 
   

   return Bsig;

};


//// synapsesTD2::init /////////////////////////////////
//

result synapsesTD2::init()
{
   
   // reading number of delays
   
   if(ReadVal((char*)DelaysKey,Delays) != OKAY)
   {
      ErrOutput << Sign << DelaysKey << SepEQU << " not specified.\n";
      return FAILED;
   }
  
   // input/output locations and sizes
   
   int i,t;
   
   N = Xin->Size[0];
   AllocArray(X, Delays+1, N);
   for (t=0; t<=Delays; t++)
      for (i=0; i<N; i++)
	 X[t][i]=0;
 
 
   V  = Yout->Data[0];
   if (Yout->Size[0] != N)
   {
      ErrOutput << Sign << "Dimensions of input ans output don't match.\n";
      return FAILED;
   }
   
   // create weights and thresholds
   
   W = NULL;   Thld = NULL;
   AllocArray(W, Delays+1, N);
   AllocArray(Thld, Delays+1, N);
   if(!W || !Thld)
      return FAILED;
      
   // load weights and thresholds
   
   if(ReadArray((char*)WeightKey, W, Delays+1, N) != OKAY)
   {
      ErrOutput << Sign << "No 2-d weight matrix specified.\n";
      return FAILED;
   }

   if(ReadArray((char*)ThldKey, Thld, Delays+1, N) != OKAY)
   {
      ErrOutput << Sign << "No 2-d matrix of thresholds specified.\n";
      return FAILED;
   }

 
   return OKAY;
   
};



//// synapsesTD2::save /////////////////////////////////
//

result synapsesTD2::save()
{
   
   WriteVal((char*)DelaysKey,Delays);WriteLn();
   WriteArray((char*)WeightKey, W, Delays+1, N);
   WriteArray((char*)ThldKey, Thld, Delays+1, N);

   return OKAY;

};   

//// synapsesTD2::forward /////////////////////////////////
//

sysig synapsesTD2::forward()
{

   int t,k;
   
   // reading new input
   
   for (k=0; k<N; k++)
      X[0][k] = Xin->Data[0][k];

   
   // calculating output
   
   for(k=0;k<N;k++)
   {
      V[k]=0.0;
      for (t=0; t <= Delays; t++)
      {
         V[k] += W[t][k] * X[t][k];
         V[k] -= Thld[t][k];
      }
   }
   
   // shifting input buffer
   
   for (t=1; t<=Delays; t++)
      for (k=0; k<N; k++)
	 X[t][k] = X[t-1][k];
 
    
   return Fsig;

};

//// synapsesTD2::backward /////////////////////////////////
//

sysig synapsesTD2::backward()
{

   return NOsig;

};


//// synapsesTDBP2::init /////////////////////////////////
//

result synapsesTDBP2::init()
{
   
   int t,k;
   
   // reading number of delays
   if(ReadVal((char*)DelaysKey,Delays) != OKAY)
   {
      ErrOutput << Sign << DelaysKey << SepEQU << " not specified.\n";
      return FAILED;
   }
      
   // input/output locations and sizes
   
   N = Xin->Size[0];
   AllocArray(X, Delays+1, N);
   for (t=0; t<=Delays; t++)
      for (k=0; k<N; k++)
	 X[t][k]=0;
   Xb = Xout->Data[0];

   if(Xout->Size[0])
      if  (N != Xout->Size[0]) 
      {
         ErrOutput << Sign << "Vectors X and Xb have diferent sizes.\n";
         return FAILED;
      }
 
   V = Yout->Data[0];
   Vb = Yin->Data[0];
   if((Yout->Size[0]) != Yin->Size[0])
   {
      ErrOutput << Sign << "Vectors V and Vb have diferent sizes.\n";
      return FAILED;
   }
   
   if(N != Yout->Size[0])
   {
      ErrOutput << Sign << "Input and output vectors have diferent sizes.\n";
      return FAILED;
   }


   
   // reading learning rate
   if(ReadVal((char*)LRateKey,LearnRate) != OKAY)
   {
      ErrOutput << Sign << LRateKey << SepEQU << " not specified.\n";
      return FAILED;
   }
      
   // create weights and thresholds
   W = NULL;   Thld = NULL;
   AllocArray(W, Delays+1, N);
   AllocArray(Thld, Delays+1, N);
   if(!W || !Thld)
      return FAILED;

   // set weights, if not given then zeros
   boolean ZeroWeights = FALSE;
   if(ReadArray((char*)WeightKey, W, Delays+1, N) != OKAY)
   {
      for (t=0; t<=Delays; t++) 
         for(k=0;k<N;k++)
            W[t][k] = 0.0;
      ZeroWeights = TRUE;
   }
   
   // set thresholds, if not given then zeros
   if(ReadArray((char*)ThldKey, Thld, Delays+1, N) != OKAY)
      for (t=0; t<=Delays; t++)
         for(k=0;k<N;k++)
     	    Thld[t][k] = 0.0;
   	
   // if some random bound is given
   //   add noise to the weights&thresholds
   // in this way the weights&thresholds can be preset in random manner
   if(ReadVal((char*)MaxRndKey, RandMax) == OKAY)
   { 
      // adding noise, DC = 0
      for (t=0; t<=Delays; t++)
         for(k=0;k<N;k++)
         {
	    W[t][k]    = 2.0*RandMax * randdouble - RandMax;
     	    Thld[t][k] = 2.0*RandMax * randdouble - RandMax;
         }
      }

   else 
      if(ZeroWeights)
         WarnOutput << Sign << "All weights are set to zero.\n";
      

   return OKAY;
      
};




//// synapsesTDBP2::save /////////////////////////////////
//

result synapsesTDBP2::save()
{
   
   WriteVal((char*)DelaysKey,Delays);WriteLn();
   WriteVal((char*)LRateKey,LearnRate);WriteLn();
   WriteArray((char*)WeightKey, W, Delays+1, N);
   WriteArray((char*)ThldKey, Thld, Delays+1, N);

   return OKAY;

}; 

//// synapsesTDBP2::forward /////////////////////////////////
//

sysig synapsesTDBP2::forward()
{

   int t,k;
   
   // reading new input
   
   for (k=0; k<N; k++)
      X[0][k] = Xin->Data[0][k];


   // calculating output
   
   for(k=0;k<N;k++)
   {
      V[k]=0.0;
      for (t=0; t <= Delays; t++)
      {
         V[k] += W[t][k] * X[t][k];
         V[k] -= Thld[t][k];
      }
   }

   // shifting input buffer should be delayed until after
   // backward function
   
      
   return Fsig;

};

//// synapsesTDBP2::backward /////////////////////////////////
//

sysig synapsesTDBP2::backward()
{

   int t,k;
   
   
   // backpropagation first, only for zero delay layer
   
   if(Xout->Size[0])
      for(k=0;k<N;k++)
         Xb[k] = W[0][k] * Vb[k];
        

   // adaptation
   
   if(LearnRate != 0.0)
      for (t=0; t<=Delays; t++)
         for(k=0;k<N;k++)
         {
	    W[t][k]    += LearnRate * Vb[k] * X[t][k];
            Thld[t][k] -= LearnRate * Vb[k];
         }
   
   // now it's safe to shift the buffer

   for (t=1; t<=Delays; t++)
      for (k=0; k<N; k++)
	 X[t][k] = X[t-1][k];
 
   

   return Bsig;

};


//// softmax::init //////////////////////////////////////
//

result softmax::init()
{
  
   // locate I/O vectors, set size
   X = Xin->Data[0];
   Y = Yout->Data[0];
   if( (N=Xin->Size[0]) != Yout->Size[0] )
   {
      ErrOutput << Sign <<"Forward inputs/outputs inconsistency.\n";
      return FAILED;
   }
   
   return OKAY;
}
   


//// softmax::forward ////////////////////////////////////
//

sysig softmax::forward()
{
   int i;
   double sum;
   
   sum = 0.0;
   for (i=0;i<N;i++)
      sum += exp(X[i]);
   
   for (i=0;i<N;i++)
      Y[i]=exp(X[i])/sum;
      
   return Fsig;

};   



//// softmax::backward ///////////////////////////////////
//

sysig softmax::backward()
{

   return Bsig;

}

//// softmaxBP::init /////////////////////////////////
//

result softmaxBP::init()
{

   // locate I/O vectors
   Xb = Xout->Data[0];
   Yb = Yin->Data[0];
   if ( (Xout->Size[0] != Yin->Size[0]) || (Xout->Size[0] != Xin->Size[0]) )
   {
      ErrOutput << Sign << "Backward inputs/outputs inconsistency.\n";
      return FAILED;
   }
   
   // the rest is the same as in softmax
   
   return softmax::init();
   
}



//// softmaxBP::backward ////////////////////////////////////
//

sysig softmaxBP::backward()
{
   
   int i,j;
   

   for(i=0;i<N;i++)
   {
      Xb[i] = Yb[i];
	 
      for (j=0;j<N;j++)
         Xb[i] -= Y[j]*Yb[j];
	 
      Xb[i] *= Y[i];
   }

   
   return Bsig;
   
};   


   

//// teacher::init /////////////////////////////////
//

result teacher::init()
{
  
   // input/output locations and sizes
       
   if(  ( (N = Xin->Size[0]) != Xin->Size[1]) || (N != Xout->Size[0]) ||
        (N != Yout->Size[0]) || (N != Yout->Size[1]) ||
        (N != Yout->Size[2])
     )
   {
      ErrOutput << Sign 
      << "Vector lengths of desired and actual responses \n"
      << "and of error don't match.\n";
      return FAILED;
   }

   if ( (Yout->Size[3] != 1) || (Yout->Size[4] != 1) )
      {
         ErrOutput << Sign 
	    << "Distance is a scalar (not a vector).\n";
	 return FAILED;
      }
   
   d = Xin->Data[0];
   X = Xin->Data[1];
   e = Xout->Data[0];
   dist = Yout->Data[3];      // distance for one training pattern
   totaldist = Yout->Data[4]; // accumulates the distances over the patterns

   *dist=0; *totaldist=0; patterns = 0;

   return OKAY;

};

//// teacher::forward /////////////////////////////////
//

sysig teacher::forward()
{

   int i;

   *dist = 0.0;

   for(i=0;i<N;i++)
   {
      Yout->Data[0][i] = d[i];
      Yout->Data[1][i] = X[i];
      Yout->Data[2][i] = e[i] = d[i] - X[i];
      *dist += e[i] * e[i];
      Yout->Data[5][i] = Xin->Data[2][i];
   }

   *totaldist += *dist;
   patterns++;

   return Fsig;

};



//// teacher::backward /////////////////////////////////
//

sysig teacher::backward()
{

   int i;

   // this is a bit luxury since error vector is effectively computed twice
   for(i=0;i<N;i++)
      e[i] = d[i] - X[i];


   return Bsig;

};

// teacher::shutdown //////////////////////////////////////
//

result teacher::shutdown()
{
   cout << "\nTraining complete.\n\n";
   cout << "Total energy       : " << *totaldist/2 << "\n";
   cout << "Patterns seen      : " << patterns << "\n";
   cout << "Energy per pattern : " << *totaldist/patterns/2 << "\n";

   return OKAY;
}


//// tester::init /////////////////////////////////
//

result tester::init()
{

   // input/output locations and sizes

   if(  ( (N = Xin->Size[0]) != Xin->Size[1]) ||
	(N != Yout->Size[0]) || (N != Yout->Size[1]) ||
	(N != Yout->Size[2])
     )
   {
      ErrOutput << Sign
      << "Vector lengths of desired and actual responses \n"
      << "and of error don't match.\n";
      return FAILED;
   }

   if ( (Yout->Size[3] != 1) || (Yout->Size[4] != 1) )
      {
	 ErrOutput << Sign
	    << "Distance is a scalar (not a vector).\n";
	 return FAILED;
      }

   d = Xin->Data[0];
   X = Xin->Data[1];
   e = Yout->Data[2];
   dist = Yout->Data[3];      // distance for one training pattern
   totaldist = Yout->Data[4]; // accumulates the distances over the patterns

   *dist=0 ; *totaldist=0; patterns=0;

   return OKAY;

};

//// tester::forward /////////////////////////////////
//

sysig tester::forward()
{

   int i;

   *dist = 0.0;

   for(i=0;i<N;i++)
   {
      Yout->Data[0][i] = d[i];
      Yout->Data[1][i] = X[i];
      e[i] = d[i] - X[i];
      *dist += e[i] * e[i];
      Yout->Data[5][i] = Xin->Data[2][i];
   }

   *totaldist += *dist;
   patterns++;

   return Fsig;

};



// tester::shutdown //////////////////////////////////////
//

result tester::shutdown()
{
   cout << "\nTesting complete.\n\n";
   cout << "Total energy       : " << *totaldist/2 << "\n";
   cout << "Patterns seen      : " << patterns << "\n";
   cout << "Energy per pattern : " << *totaldist/patterns/2 << "\n";

   return OKAY;
}

//// confusiondummy::init /////////////////////////////////
//

result confusiondummy::init()
{
   ConfTabName = NULL; ConfMatName = NULL; N =0;

   ReadVal((char*)ConfTabKey, ConfTabName);
   ReadVal((char*)ConfMatKey, ConfMatName);
   ReadVal((char*)CodeSizeKey, N);

   return OKAY;
}

//// confusiondummy::save /////////////////////////////////
//

result confusiondummy::save()
{

   WriteVal((char*)ConfTabKey, ConfTabName); WriteLn();
   WriteVal((char*)ConfMatKey, ConfMatName); WriteLn();
   WriteVal((char*)CodeSizeKey, N); WriteLn();

   return OKAY;

};
