//////////////////////////////////////////////////////////////////////////////
//
// File:	neural2.h
//
//////////////////////////////////////////////////////////////////////////////

#ifndef NEURAL2_H
#define NEURAL2_H

#include "black.h"
#include <math.h>


// boxes--type identifiers in neural2.C

extern const char *SynapsesTD;    
extern const char *SynapsesTDBP; 
extern const char *SynapsesTD2;
extern const char *SynapsesTDBP2;
extern const char *SoftMax;
extern const char *SoftMaxBP;
extern const char *Teacher;
extern const char *Tester;
extern const char *ConfusionDummy;
extern const char *ConfTabKey;
extern const char *ConfMatKey;
extern const char *CodeSizeKey;

// key words

extern const char *DelaysKey;


//////////////////////////////////////////////////////////////////////////////
//
// CLASS: SynapsesTD
//
//////////////////////////////////////////////////////////////////////////////

// synaptic layer for time delayed neural networks (TDNN)

class synapsesTD : public blackBox
{
   
   protected: 
      int	Delays; // number of delays, zero means standard synapses
      int	Nx, Nv;
      signal	**X, *V; // X<time (row) index t, vector (column) index i>
      double	***W;
           // indexing weights: "1 output neuron per line per time block"
      	   // W<time (block) index t, output (row) index k,input (column) index j>
      	   //   = W[t][k][j]
      	   //  i.e. #rows:	Nv	#columns: Nx
      	   // t=0 means current input, t=1 previous input etc.
      
      double	**Thld;	// Thld<time (row) index t, vector (column) index i>
    
   public:
     synapsesTD(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:
      virtual result init();
      virtual result save();
      virtual sysig  forward();
      virtual sysig  backward();
      
};


//////////////////////////////////////////////////////////////////////////////
//
// CLASS: synapsesTDBP
//
//////////////////////////////////////////////////////////////////////////////

// 'synapsesTDBP' has 1 connection on all the tracks | inputs | outputs
//   -it computes the synaptic potencial based on weights and thresholds
//     as inherited from synapses
//   -backward propagation of the gradient as well as adaptation is possible
//     in the basic form of backpropagation algorithm
//

class synapsesTDBP : public synapsesTD
{
   
   protected:
      double	RandMax,LearnRate;
      signal	*Xb, *Vb;

      
   public:
     synapsesTDBP(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: synapsesTD(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig forward();
      virtual sysig backward();

};


//////////////////////////////////////////////////////////////////////////////
//
// CLASS: SynapsesTD2
//
//////////////////////////////////////////////////////////////////////////////

// Same as SynapsesTD, but somehow restricted.
// Number of inputs and outputs should be the same and no interconnections between
// input and output neurons with different indexes are allowed.
// Uses up less memory and computation time.


class synapsesTD2 : public blackBox
{
   
   protected: 
      int	Delays; // number of delays, zero means standard synapses
      int	N;
      signal	**X, *V; // X<time (row) index t, vector (column) index i>
      double	**W;
           // indexing weights: "1 time slot per line"
      	   // W<time (row) index t, neuron (column) index k>
      	   //   = W[t][k]
      	   // t=0 means current input, t=1 previous input etc.
      
      double	**Thld;	// Thld<time (row) index t, vector (column) index i>
    
   public:
     synapsesTD2(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:
      virtual result init();
      virtual result save();
      virtual sysig  forward();
      virtual sysig  backward();
      
};

//////////////////////////////////////////////////////////////////////////////
//
// CLASS: synapsesTDBP2
//
//////////////////////////////////////////////////////////////////////////////

class synapsesTDBP2 : public synapsesTD2
{
   
   protected:
      double	RandMax,LearnRate;
      signal	*Xb, *Vb;

      
   public:
     synapsesTDBP2(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: synapsesTD2(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
      virtual sysig forward();
      virtual sysig backward();

};

//////////////////////////////////////////////////////////////////////////////
//
// CLASS: softmax
//
//////////////////////////////////////////////////////////////////////////////

// softmax non linearity
// outputs are normalised so that sum equals one

class softmax : public blackBox
{
   protected:
      double *X,*Y;
      int N;
   
   public:
     softmax(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:
      virtual result init();
      virtual sysig forward();
      virtual sysig backward();

};

//////////////////////////////////////////////////////////////////////////////
//
// CLASS: softmaxBP
//
//////////////////////////////////////////////////////////////////////////////

class softmaxBP : public softmax
{
   
   protected:
      double *Xb,*Yb;
       
   public:
     softmaxBP(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: softmax(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:
     virtual result init();
     virtual sysig backward();

};


//////////////////////////////////////////////////////////////////////////////
//
// CLASS: teacher
//
//////////////////////////////////////////////////////////////////////////////

//  Xin  : port 0 : desired pattern
//         port 1 : actual pattern
//         port 3 : input pattern
//  Xout : port 0 : error vector
//  Yout : port 0 : desired pattern
//         port 1 : actual pattern
//         port 2 : error vector
//         port 3 : distance of the error vector
//         port 4 : distance accumulated over the patterns
//         port 5 : input pattern
//
//  forward method always returns Fsig, so the teacher has to be followed
//     by e.g. VectorOut to revert the signal flow. 
//  backward method computes the error and returns Bsig

class teacher : public blackBox
{
   
   protected:
      signal	*X,  *d,  *e, *dist, *totaldist;
      int 	N;
      long int  patterns;
      
   public:
      teacher(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual sysig forward();
      virtual sysig backward();
      virtual result shutdown();

};

//////////////////////////////////////////////////////////////////////////////
//
// CLASS: tester
//
//////////////////////////////////////////////////////////////////////////////

//  Xin  : port 0 : desired pattern
//         port 1 : actual pattern
//         port 3 : input pattern
//  
//  Yout : port 0 : desired pattern
//         port 1 : actual pattern
//         port 2 : error vector
//         port 3 : distance of the error vector
//         port 4 : distance accumulated over the patterns
//         port 5 : input pattern


class tester : public blackBox
{
   
   protected:
      signal	*X,  *d,  *e, *dist, *totaldist;
      int 	N;
      long int  patterns;
      
   public:
      tester(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:
      virtual result init(); 
      virtual sysig forward();
      virtual result shutdown();

};


//////////////////////////////////////////////////////////////////////////////
//
// CLASS: confusiondummy
//
//////////////////////////////////////////////////////////////////////////////

class confusiondummy : public blackBox
{
   
   protected:
      int	X, D;
      int 	N;
       
      FILE	*ConfTabFile, *ConfMatFile;
      char	*ConfTabName, *ConfMatName;
      
   public:
     confusiondummy(char *TypeStr, char *SignatStr, int Xins, int Youts, int Yins, int Xouts)
	: blackBox(TypeStr, SignatStr, Xins, Youts, Yins, Xouts) {};
	      	 
   protected:      
      virtual result init();
      virtual result save();
 
};

#endif
